#	Copyright (c) Alexander Sedov 2008

#	This file is part of Nodes.
#	
#	Nodes is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	Nodes is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with Nodes.  If not, see <http://www.gnu.org/licenses/>.

"""Core of Nodes.
"""
import copy
import cnodes #here
from numpy import*
flotype=float32 #type of NumPy arrays
class Node(object):
	"""Base class."""
	def __init__(self, inputs, outputs, iweights=None, oweights=None):
		"""Initialization.
		inputs, outputs -- list of get/set interfaces to other Nodes (Synaps subclasses)
		iweights, oweights -- lists/arrays of initial weights (default to ones). """
		self.input_weights=iweights
		self.output_weights=oweights
		self.inputs=inputs
		self.outputs=outputs
		self.init_prepare()
		self.check_valid()
		self.free=True

	def touch(self):
		"""Touching Node for calculations.
		Whenever applied, cause touches of whole net."""
		self.prepare_calc()
		self.copy_weights()
		self.get_values()
		self.mul_iweights()
		self.preprocess()
		self.calculate()
		self.teach()
		self.postprocess()
		self.mul_oweights()
		self.put_values()
		self.dump_weights()
		self.fix_calc()

	def go_fwd(self, exclude=None):
		"""Going forward.
		ivalues must be setted by others."""
		self.prepare_calc()
		self.copy_weights()
		if exclude is not None:
			self.get_values(exclude)
		self.mul_iweights()
		self.preprocess()
		self.calculate()
		self.teach()
		self.postprocess()
		self.mul_oweights()
		self.put_values()
		self.dump_weights()
		self.fix_calc()

	def go_bwd(self, exclude=None):
		"""Going backward.
		ovalues will be getted by others."""
		self.prepare_calc()
		self.copy_weights()
		self.get_values()
		self.mul_iweights()
		self.preprocess()
		self.calculate()
		self.teach()
		self.postprocess()
		self.mul_oweights()
		if exclude is not None:
			self.put_values(exclude)
		self.dump_weights()
		self.fix_calc()

	def prepare_calc(self):
		pass

	def fix_calc(self):
		pass

	def copy_weights(self):
		"""Copying weights for backuping
		and multi-access support."""
		self.iweights=array(self.input_weights)
		self.oweights=array(self.output_weights)

	def dump_weights(self):
		"""Dumping weights into."""
		self.input_weights=self.iweights
		self.output_weights=self.oweights
		del self.iweights, self.oweights

	def init_prepare(self):
		"""Finishing initialization."""
		self.iN=len(self.inputs)
		self.oN=len(self.outputs)
		#Creating arrays of values and (if needed) weights.
		self.ivalues=zeros(self.iN, flotype)
		self.ovalues=zeros(self.oN, flotype)
		if self.input_weights is None:
			self.input_weights=ones(self.iN, flotype)
		else:
			self.input_weights=array(self.input_weights, flotype)
		if self.output_weights is None:
			self.output_weights=ones(self.oN, flotype)
		else:
			self.output_weights=array(self.output_weights, flotype)
		self.input=array(self.ivalues)
		self.output=array(self.ovalues) #С чем оперирует calculate
		#Создаём массивы-истории значений(для прямого распространения)
		self.ilog=array(self.ivalues)
		self.olog=array(self.ovalues)
		self.N=0	#Сколько раз провели обучение.
		self.velocity=1 #Скорость обучения, точнее, её множитель.

	def check_valid(self):
		"""."""
		assert len(self.input_weights)==self.iN
		assert len(self.output_weights)==self.oN
		return True

	def get_values(self, exclude=None):
		"""Reading values into ivalues[]."""
		self.free=False
		for i in range(self.iN): #it can be impl with map(), but without :exclude support
			if i==exclude: continue
			val=self.inputs[i].get()
			if val is not None:
				self.ivalues[i]=val
		self.free=True
	def mul_iweights(self):
		self.free=False
		self.input=self.ivalues*self.iweights #good thing - NumPy arrays!
		self.free=True

	def mul_oweights(self):
		self.ovalues=self.output*self.oweights

	def put_values(self, exclude=None):
		"""Writing values from ovalues[]."""
		for j in range(self.oN):
			if j==exclude: continue
			self.outputs[j].put(self.ovalues[j])

	def preprocess(self):
		"""Preparing to calculations."""
		self.output=self.output*0. #обнуляем

	def postprocess(self):
		"""Handling calculated."""
		self.input=self.input*0.

	def calculate(self):
		"""Calculations.
		before: input[] filled.
		after: output[] filled."""
		raise NotImplementedError, "calculations must be implemented in subclasses."

	def teach(self):
		"""Teaching (weights correction)."""
		self.forth_propagation()
		#self.back_propagation() # require :targets

	def forth_propagation(self):
		"""Forward propagation (part of conception).
		"""
		iweights=array(self.iweights)
		oweights=array(self.oweights)
		#Stage #1. Factors calculations.
		for a in xrange(self.iN):
			x=self.ivalues[a]
			y=self.ilog[a]
			pow=cnodes.dsigmoid(x-y)*4 	#Believing power of new value.
						#even function, min 0, max 1.
			k1=self.N
			k2=pow*self.velocity
			val=(k1*y+k2*x)/(k1+k2)
			self.ilog[a]=val
			km=cnodes.dsigmoid(y-val)*self.velocity #Modifying coefficient.
			iweights[a]=iweights[a]*(1-km)
		for a in xrange(self.oN):
			x=self.ovalues[a]
			y=self.olog[a]
			pow=cnodes.dsigmoid(x-y)*4
			k1=self.N
			k2=pow*self.velocity
			val=(k1*y+k2*x)/(k1+k2)
			self.olog[a]=val
			km=cnodes.dsigmoid(y-val)*self.velocity
			oweights[a]=oweights[a]*(1-km)
		
		#Stage #2. Ajusting.
		for a in xrange(self.iN):
			val, log=self.ivalues[a], self.ilog[a]
			if log:
				w=val/log
				wn=iweights[a]
				add=(w-wn)*self.velocity*cnodes.dsigmoid(w-wn)*4
				iweights[a]=iweights[a]+add
			else:
				self.ilog[a]=val
		for a in xrange(self.oN):
			val, log=self.ovalues[a], self.olog[a]
			if log:
				w=val/log
				wn=oweights[a]
				add=(w-wn)*self.velocity*cnodes.dsigmoid(w-wn)*4
				oweights[a]=oweights[a]+add
			else: self.olog[a]=val
		
		#Stage #3. Calculating of keyfactors.
		for a in xrange(self.iN):
			x=self.ivalues[a]
			key=cnodes.dblsigmoid(x)*cnodes.dsigmoid(x) #`keyness' of value.
			iweights[a]=iweights[a]*(1+(key*self.velocity))
		for a in xrange(self.oN):
			x=self.ovalues[a]
			key=cnodes.dblsigmoid(x)*cnodes.dsigmoid(x)
			oweights[a]=oweights[a]*(1+(key*self.velocity))
		
		#Stage #4. Merging.
		for a in xrange(self.iN):
			x=iweights[a]
			y=self.iweights[a]
			stbl=abs(cnodes.dmsigmoid(y)) #Unstability of weight(0 is very stable).
			force=cnodes.dsigmoid(x) #Force factor.
			k1=self.N
			k2=stbl+force #Teaching velocity already applied.
			self.iweights[a]=(k1*y+k2*x)/(k1+k2)
		for a in xrange(self.oN):
			x=oweights[a]
			y=self.oweights[a]
			stbl=abs(cnodes.dmsigmoid(y))
			force=cnodes.dsigmoid(x)
			k1=self.N
			k2=stbl+force
			self.oweights[a]=(k1*y+k2*x)/(k1+k2)
		self.N+=1

	def back_propagation(self, targets):
		"""Back propagation.
		Described in many articles."""
		assert len(targets)==self.oN
		targets=array(targets)
		delta=(targets-self.ovalues)/self.output_weights*array(map(self.calc_drv, self.ovalues))
		d=sum(delta)
		self.input_weights+=(d*self.ivalues)*self.velocity
		map(lambda x, y: x.back_propagate(d*y), self.inputs, self.input_weights) #обратно распространяем
	def calc_drv(self, val):
		"""Turning f(x) into f'(x) (not x to f'(x))!"""
		raise NotImplementedError, "derivative calculation depends on calculations."
	
	#Pickling :-)
	def __getstate__(self):
		return self.inputs, self.outputs, self.iweights, self.oweights, self.ilog, self.olog,\
			self.__getstate_extra__()

	def __setstate__(self, state):
		self.inputs, self.outputs, self.iweights, self.oweights, self.ilog, self.olog, extra_state=state
		self.iN=len(self.inputs)
		self.oN=len(self.outputs)
		self.ivalues=zeros(self.iN)
		self.ovalues=zeros(self.oN)
		self.__setstate_extra__(extra_state)

	def __getstate_extra__(self):
		return None

	def __setstate_extra__(self, state):
		pass

#Synapses (accesors)
class Synaps(object):
	def __init__(self, node, index):
		self.node=node
		self.index=index

class InputSynaps(Synaps):
	def __init__(self, node, index):
		Synaps.__init__(self, node, index)
		self.cache=None

	def get(self):
		try:
			x=self.node.ovalues[self.index]
		except:
			if self.cache:
				x=self.cache
			self.node.go_bwd(self.index)
			try:
				x=self.node.ovalues[self.index]
			except:
				pass
		else:
			if self.cache is None:
				self.cache=x
			elif self.cache==x: #no changes
				self.node.go_bwd(self.index) #excluding self
				self.cache=x=self.node.ovalues[self.index]
			return x

		def back_propagate(self, value):
			try:
				l=self.node.ovalues[:]
			except:
				return
			else:
				try:
					l[self.index]=value
				except:
					return
				else:
					self.node.back_propagate(l)

class OutputSynaps(Synaps):
	def __init__(self, node, index):
		Synaps.__init__(self, node, index)
		self.retry=None

	def put(self, value):
		if self.retry is not None:
			try:
				self.node.ivalues[self.index]=self.retry
			except:
				pass
			else:
				self.node.go_fwd(self.index)
				self.retry=None
		try:
			x=self.node.ivalues[self.index]
		except:
			x=None #n-equal to all. Like NaN.
		if x!=value or self.node.free:
			try:
				self.node.ivalues[self.index]=value
			except:
				self.retry=value
			else:
				self.node.go_fwd(self.index)
